﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace Unclassified
{
	// Info: http://www.radsoftware.com.au/articles/clipboardmonitor.aspx
	static class ClipboardWatcher
	{
		#region Win32 interop

		[DllImport("User32.dll")]
		private static extern IntPtr SetClipboardViewer(IntPtr hWnd);

		[DllImport("User32.dll")]
		private static extern bool ChangeClipboardChain(
			IntPtr hWndRemove,  // handle to window to remove
			IntPtr hWndNewNext  // handle to next window
			);

		[DllImport("user32.dll")]
		private static extern int SendMessage(IntPtr hwnd, int wMsg, IntPtr wParam, IntPtr lParam);

		private const int WM_DRAWCLIPBOARD = 0x0308;
		private const int WM_CHANGECBCHAIN = 0x030D;

		#endregion Win32 interop

		#region Static members

		private static int refCount;
		private static Window window;

		public static event EventHandler Changed
		{
			add
			{
				refCount++;
				if (refCount > 0)
				{
					window = new Window();
				}
				window.ClipboardChanged += value;
			}
			remove
			{
				window.ClipboardChanged -= value;
				if (refCount > 0)
				{
					refCount--;
				}
				if (refCount == 0)
				{
					window.DestroyHandle();
					window = null;
				}
			}
		}

		#endregion Static members

		#region Window management

		class Window : NativeWindow
		{
			private IntPtr nextClipboardViewer;

			public EventHandler ClipboardChanged;

			public Window()
			{
				CreateParams cp = new CreateParams();
				CreateHandle(cp);
			}

			protected void OnClipboardChanged()
			{
				EventHandler handler = ClipboardChanged;
				if (handler != null)
				{
					handler(this, EventArgs.Empty);
				}
			}

			protected override void OnHandleChange()
			{
				if (Handle != IntPtr.Zero)
				{
					nextClipboardViewer = SetClipboardViewer(Handle);
				}
			}

			public override void DestroyHandle()
			{
				ChangeClipboardChain(this.Handle, nextClipboardViewer);
				base.DestroyHandle();
			}

			protected override void WndProc(ref Message m)
			{
				switch (m.Msg)
				{
					case WM_DRAWCLIPBOARD:
						OnClipboardChanged();
						SendMessage(nextClipboardViewer, m.Msg, m.WParam, m.LParam);
						break;

					case WM_CHANGECBCHAIN:
						if (m.WParam == nextClipboardViewer)
						{
							nextClipboardViewer = m.LParam;
						}
						else
						{
							SendMessage(nextClipboardViewer, m.Msg, m.WParam, m.LParam);
						}
						break;

					default:
						base.WndProc(ref m);
						break;
				}
			}
		}

		#endregion Window management
	}
}
